"""Transformers working on NLTK tree objects."""
from __future__ import annotations

from typing import Any, Iterable, List, Union

import nltk
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


class StringToTreeTransformer(BaseEstimator, TransformerMixin):
    """
    Parses tree representations into their nltk.Tree object form.

    input: list of list of strings
    output: list of list of nltk.Tree objects
    """

    def fit(self, _x: Any, _y: Any = None) -> StringToTreeTransformer:
        """Fit the model."""
        return self

    def transform(self, x: List[List[str]], _y: Any = None) -> List[List[nltk.Tree]]:
        """Transform the data."""
        result = []
        for document in x:
            ret = []
            if isinstance(document, str):
                raise Exception(
                    "this transformer only takes document which "
                    "are already split into sentences. Please use a sentence "
                    "splitter before (e.g., by using stanfordnlp)"
                )
            for line in document:
                if line and line.strip():
                    tree = nltk.Tree.fromstring(line)
                    ret.append(tree)
            result.append(ret)
        return result


class WordToPosTransformer(BaseEstimator, TransformerMixin):
    """
    Creates a POS-only representation of a tree sentence.

    input: list of list of trees
    output: list of list of strings
    """

    def fit(self, _x: Any, _y: Any = None) -> WordToPosTransformer:
        """Fit the model."""
        return self

    def transform(self, x: List[List[nltk.Tree]], _y: Any = None) -> List[List[str]]:
        """Transform the data."""
        ret = []
        for document in x:
            tmp = []
            for tree in document:
                tmp += [x[1] for x in tree.pos()]
            ret.append(tmp)
        return ret


class TreeChainTransformer(BaseEstimator, TransformerMixin):
    """
    Create chains of strings with an optional max length from tree structures.

    For each leaf node, its path to the root node is called a chain.
    This transformer returns every subchain with a certain length from every
    chain. Common ancester subchains are only returned once.
    For example, in the tree (1 (2 (3, 4))) and max_lenth=2, the subchain 1-2
    is only returned once, although it is part of both chains 1-2-3 and 1-2-4.

    input document: list of nltk.Tree objects (one for each sentence)

    output document: the format of the output documents depend on the
    combination parameters.




    """

    def __init__(
        self,
        max_length: int = None,
        combine_chain_elements: str = None,
        combine_chains: str = None,
        combine_strings: str = None,
    ):
        """
        Initialize class.

        Args:
            max_length (int): the maximum length of the produced chains.
                If set to None, full chains of any length will be returned.

        combine_chain_elements (str): if set to any string other than None,
                each chain will be returned as a string
                (e.g., "PRON VERB NOUN")
                instead of a list. The parameter is used to join the chains.
                Defaults to None.

        combine_chains (str|None):  if set to true, return each document as the
                concatenation of all chains for all trees in the document. If
                set to None, returns a list of sentences for each document,
                where each sentence consists of a list of chains. The parameter
                 is used to join the chains. Defaults to None.

        combine_strings (str|None): if set, it will be used to join the tree
                chains together to a single string per document. The parameter
                is used to join the chains. Defaults to None.
        """
        if combine_strings is not None and combine_chains is None:
            raise ValueError(
                "if combine_strings is set, both combine_chains and "
                "combine_chain_elementsmust be set too."
            )
        if combine_chains is not None and combine_chain_elements is None:
            raise ValueError(
                "if combine_chains is set, combine_chain_elements must be set " "too."
            )

        self.max_length = max_length
        self.combine_chain_elements = combine_chain_elements
        self.combine_chains = combine_chains
        self.combine_strings = combine_strings

    def fit(self, _x: Any, _y: Any = None) -> TreeChainTransformer:
        """Fit the model."""
        return self

    # please be careful, it is dangerous in here!
    def _find_chains(self, tree: nltk.Tree, prev_labels: List[str]) -> List[List[str]]:
        ret_full = []
        if isinstance(tree, str) or len(tree) == 0:
            # the leaf nodes of nltk trees are strings
            if not self.max_length or len(prev_labels) < self.max_length:
                ret_full.append(prev_labels[:])
            return ret_full
        if self.max_length and len(prev_labels) == self.max_length:
            # case: intermediary subchain that stops before the root
            prev_labels = prev_labels[1:]

        prev_labels.append(tree.label())
        if len(prev_labels) == self.max_length:
            # by adding this node, we have completed a subchain
            ret_full.append(prev_labels[:])
        for child in tree:
            ret_full += self._find_chains(child, prev_labels[:])
        return ret_full

    def transform(
        self, documents: List[List[nltk.Tree]], _y: Any = None
    ) -> Union[List[str], List[List[str]]]:
        """Transform the data."""
        result = []
        for document in documents:
            #  a document consists of a list of nltk.Tree objects
            #  one tree represents one sentence.
            new_document: Any = []
            for tree in document:
                #  every tree is split into a list of chains, whereas
                #  every chain is a list of symbols.
                chains: Any = self._find_chains(tree, [])

                if self.combine_chain_elements is not None:
                    chains = [self.combine_chain_elements.join(x) for x in chains]
                if self.combine_chains is not None:
                    new_document.append(self.combine_chains.join(chains))
                else:
                    new_document.append(chains)

            if self.combine_strings is not None:
                new_document = self.combine_strings.join(new_document)
            result.append(new_document)
        return result


def _get_average_height(document: List[nltk.Tree]) -> np.array:
    """Calculate the average height of all trees in a document."""
    return np.average([tree.height() for tree in document])


def _calculate_average_children(tree: nltk.Tree) -> List[int]:
    result = []
    if type(tree) != nltk.tree.Tree or not tree:
        result.append(0)
    else:
        result.append(len(tree))
        for child in tree:
            result += _calculate_average_children(child)
    return result


def _get_average_children(document: List[nltk.Tree]) -> np.array:
    result = []
    for tree in document:
        result += _calculate_average_children(tree)
    return np.average(result)


def _get_average_inner_to_leaf_ratio(document: List[nltk.Tree]) -> np.array:
    result = []
    for tree in document:
        num_leaves = len(tree.leaves())
        num_total = len(tree.treepositions()) - 1  # root node
        result.append(num_leaves / num_total)
    return np.average(result)


def _get_max_tree_width(tree: nltk.Tree) -> int:
    maximum = 0
    if type(tree) == nltk.tree.Tree:
        for child in tree:
            maximum = max(maximum, _get_max_tree_width(child), len(tree))
    return maximum


def _get_max_child_width(document: List[nltk.Tree]) -> np.array:
    result = []
    for tree in document:
        result.append(_get_max_tree_width(tree))
    return np.average(result)


class TreeStatsVectorizer(TransformerMixin, BaseEstimator):
    """
    Calculate high-level, low-complexity tree features.

    This transformer creates aggregated high-level features that are not
    dependent on the content of the tree.

    input: list of list of trees
    output: list of list of numbers/floats
    """

    def fit(self, _x: Any, _y: Any = None) -> TreeStatsVectorizer:
        """Fit the model."""
        return self

    def transform(self, x: List[List[nltk.Tree]], _y: Any = None) -> List[List[float]]:
        """Transform the data."""
        result = []
        for document in x:
            result.append(
                [
                    _get_average_height(document),
                    _get_average_children(document),
                    _get_average_inner_to_leaf_ratio(document),
                    _get_max_child_width(document),
                ]
            )
        return result


class TreeToStringTransformer(BaseEstimator, TransformerMixin):
    """
    Transforms nltk trees into strings that can be parsed later.

    input: list of list of trees
    output: list of list of strings
    """

    def fit(self, _x: Any, _y: Any = None) -> TreeToStringTransformer:
        """Fit the model."""
        return self

    def transform(self, x: List[List[nltk.Tree]], _y: Any = None) -> List[List[str]]:
        """Transform the data."""
        ret = []
        for document in x:
            doc = []
            for tree in document:
                doc.append(tree.pformat(margin=1000000))
            ret.append(doc)
        return ret


class TreeToFlatStringTransformer(TransformerMixin, BaseEstimator):
    """
    Transforms NLTK trees into Strings by traversing the tree post-order.

    Input: Every document is a list of NLTK trees, one representing each
        sentence.
    Output: Every document is a single string.
    """

    def fit(self, _x: Any, _y: Any = None) -> TreeToFlatStringTransformer:
        """Fit the model."""
        return self

    def parse(self, tree: nltk.Tree) -> str:
        """Convert a string to a NLTK tree."""
        result = [tree.label()]
        result += [self.parse(child) for child in tree]
        return " ".join(result)

    def transform(self, documents: List[List[nltk.Tree]]) -> List[str]:
        """Transform the data."""
        result = []
        for document in documents:
            parses = [self.parse(sentence) for sentence in document]
            result.append(". ".join(parses))
        return result



"""Transformers calculating string kernels."""
from __future__ import annotations

import logging
from collections import defaultdict
from typing import Any, Callable, Dict, List, Tuple, Generator

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

logger = logging.getLogger(__name__)


def iterate_ngrams(
    string: str, ngram_min: int, ngram_max: int
) -> Generator[str, None, None]:
    for ngram_size in range(ngram_min, ngram_max + 1):
        for index in range(len(string) - ngram_size + 1):
            yield string[index : index + ngram_size].lower()


def get_all_ngram_counts(string: str, ngram_min: int, ngram_max: int) -> Dict[str, int]:
    result: Dict[str, int] = defaultdict(int)
    for ngram in iterate_ngrams(string, ngram_min, ngram_max):
        result[ngram] += 1
    return result


def presence_kernel(ngram_min: int, ngram_max: int) -> Callable:
    """
    Calculate the presence kernel, Ionescu & Popescu 2017.

    the result is a matrix, where each point [x, y] is the number of different n-grams
    that both documents x and y share.
    """

    def internal_presence_kernel(x: np.array, y: np.array) -> np.array:
        result = np.zeros((len(x), len(y)), dtype=int)
        x_counts = [get_all_ngram_counts(d, ngram_min, ngram_max) for d in x]
        y_counts = [get_all_ngram_counts(d, ngram_min, ngram_max) for d in y]
        for i, xc in enumerate(x_counts):
            xkeys = set(xc.keys())
            for j, yc in enumerate(y_counts):
                ykeys = set(yc.keys())
                result[i, j] = len(xkeys.intersection(ykeys))
        return result

    return internal_presence_kernel


def legacy_presence_kernel(ngram_min: int, ngram_max: int) -> Callable:
    """Old implementation. Slower."""

    def internal_legacy_presence_kernel(x: np.array, y: np.array) -> np.array:
        result = np.zeros((len(x), len(y)), dtype=int)
        for i, string in enumerate(x):
            ngrams1 = set()
            for ngram in iterate_ngrams(string, ngram_min, ngram_max):
                ngrams1.add(ngram)
            for j, counterstring in enumerate(y):
                ngrams2 = set()
                for ngram in iterate_ngrams(counterstring, ngram_min, ngram_max):
                    ngrams2.add(ngram)
                result[i, j] = len(ngrams1.intersection(ngrams2))
        return result

    return internal_legacy_presence_kernel


def spectrum_kernel(ngram_min: int, ngram_max: int) -> Callable:
    """Calculate the spectrum kernel, Ionescu & Popescu 2017."""

    def internal_spectrum_kernel(x: np.array, y: np.array) -> np.array:
        result = np.zeros((len(x), len(y)), dtype=int)
        x_counts = [get_all_ngram_counts(d, ngram_min, ngram_max) for d in x]
        y_counts = [get_all_ngram_counts(d, ngram_min, ngram_max) for d in y]
        for i, xc in enumerate(x_counts):
            xkeys = set(xc.keys())
            for j, yc in enumerate(y_counts):
                ykeys = set(yc.keys())
                all_ngrams = set(xkeys).intersection(set(ykeys))
                result[i, j] = sum([xc[ngram] * yc[ngram] for ngram in all_ngrams])
        return result

    return internal_spectrum_kernel


def legacy_spectrum_kernel(ngram_min: int, ngram_max: int) -> Callable:
    """Old implementation. Slower."""

    def internal_legacy_spectrum_kernel(x: np.array, y: np.array) -> np.array:
        result = np.zeros((len(x), len(y)), dtype=int)
        for i, string in enumerate(x):
            ngrams: Dict[str, int] = defaultdict(int)
            for ngram in iterate_ngrams(string, ngram_min, ngram_max):
                ngrams[ngram] += 1
            for j, counterstring in enumerate(y):
                for ngram in iterate_ngrams(counterstring, ngram_min, ngram_max):
                    result[i, j] += ngrams[ngram]
        return result

    return internal_legacy_spectrum_kernel


def intersection_kernel(ngram_min: int, ngram_max: int) -> Callable:
    """Calculate the intersection kernel, Ionescu & Popescu 2017."""

    def internal_intersection_kernel(x: np.array, y: np.array) -> np.array:
        result = np.zeros((len(x), len(y)), dtype=int)
        x_counts = [get_all_ngram_counts(d, ngram_min, ngram_max) for d in x]
        y_counts = [get_all_ngram_counts(d, ngram_min, ngram_max) for d in y]
        for i, xc in enumerate(x_counts):
            xkeys = set(xc.keys())
            for j, yc in enumerate(y_counts):
                ykeys = set(yc.keys())
                common_ngrams = set(xkeys).intersection(set(ykeys))
                result[i, j] = sum(
                    [min(xc[ngram], yc[ngram]) for ngram in common_ngrams]
                )
        return result

    return internal_intersection_kernel


def legacy_intersection_kernel(ngram_min: int, ngram_max: int) -> Callable:
    """Old implementation. Slower."""

    def internal_legacy_intersection_kernel(x: np.array, y: np.array) -> np.array:
        result = np.zeros((len(x), len(y)), dtype=int)
        for i, string in enumerate(x):
            ngrams: Dict[str, int] = defaultdict(int)
            for ngram in iterate_ngrams(string, ngram_min, ngram_max):
                ngrams[ngram] += 1
            for j, counterstring in enumerate(y):
                ngrams2 = dict(ngrams)
                for ngram in iterate_ngrams(counterstring, ngram_min, ngram_max):
                    if ngram in ngrams2 and ngrams2[ngram] > 0:
                        result[i, j] += 1
                        ngrams2[ngram] -= 1
        return result

    return internal_legacy_intersection_kernel


kernel_map = {
    "presence": presence_kernel,
    "spectrum": spectrum_kernel,
    "intersection": intersection_kernel,
}


class StringKernelTransformer(BaseEstimator, TransformerMixin):
    """
    DEPRECATED: you should probably use the Scikit SVC instead.

    Converts (string) documents to a similarity matrix (kernel).

    Input (fit): List of m strings
    Input (transform): List of n strings
    Output: m x n matrix containing the kernel similarities between the strings
    """

    def __init__(
        self,
        kernel_type: str = "intersection",
        ngram_range: Tuple[int, int] = None,
        normalize: bool = True,
    ):
        """
        Initialize the model.

        Args:
            kernel_type: one of 'intersection', 'spectrum' or 'presence'.
            ngram_range: range of n_grams to include
            normalize: whether to normalize the output across the corpus.
        """
        if ngram_range is None:
            ngram_range = (5, 10)
        self.ngram_range = ngram_range
        self.normalize = normalize
        self.kernel_type = kernel_type
        if kernel_type not in kernel_map:
            raise ValueError(f"unknown kernel: {kernel_type}")
        self._kernel = kernel_map[kernel_type]
        self._train_data = None
        self._train_kernel: np.array = None

    def fit(self, X: List[str], _y: Any = None) -> StringKernelTransformer:
        """Fit model."""
        self._train_data = np.array(X)
        self._train_kernel = self._kernel(*self.ngram_range)(X, X)
        for i in range(len(X)):
            if self._train_kernel[i, i] == 0:
                logger.error(f"zeros in diagonal at ({i},{i}) for {X[i]}")
        return self

    def transform(self, X: List[str], _y: Any = None) -> np.array:
        """Transform data."""
        X = np.array(X)
        _st = self._kernel(*self.ngram_range)(X, self._train_data)

        if not self.normalize:
            return _st

        _ss = self._train_kernel
        _tt = self._kernel(*self.ngram_range)(X, X)

        result = np.copy(_st)
        for i in range(_st.shape[0]):
            if _tt[i, i] == 0:
                logger.error(f"zeros in diagonal at ({i},{i}) for {X[i]}")
            for j in range(_st.shape[1]):
                if _tt[i, i] != 0 and _ss[j, j] != 0:
                    result[i, j] /= np.sqrt(_tt[i, i] * _ss[j, j])

        return result



"""Basic CNN model."""
import math
from typing import List, Tuple

import torch
import torch.nn as nn


def compute_first_fc_layer_input_size(
    conv_configs: List[Tuple[int, int]], n: int
) -> int:
    """
    Calculates the input dimension of the first fully connected layer.

    See https://datascience.stackexchange.com/a/40991/9281

    Args:
        conv_configs: Configurations of the convolution or max-pool layer, each a tuple
            with (kernel_size, stride) values.
        n: starting value (max sequence length)

    Returns:
        the dimension of the first fc layer
    """

    def get_output_dim(in_size: int, kernel: int, stride: int) -> int:
        return math.floor(((in_size - kernel) / stride) + 1)

    for config in conv_configs:
        n = get_output_dim(n, config[0], config[1])

    return n


class CharCNN(nn.Module):
    """Basic CNN model that can be built with variable amounts of layers etc."""

    def __init__(
        self,
        n_classes: int,
        max_seq_len: int,
        embedding_dim: int = 300,
        num_features: int = 10_000,
    ):
        super().__init__()
        self.emb_layer = nn.Embedding(num_features + 1, embedding_dim)
        self.conv_layers = nn.Sequential(
            nn.Conv1d(embedding_dim, 50, (7,), (1,)),
            nn.ReLU(),
            nn.MaxPool1d(3, 3),
            nn.Conv1d(50, 50, (5,), (1,)),
            nn.ReLU(),
            nn.MaxPool1d(3, 3),
        )
        self.fc_layers = nn.Sequential(
            nn.Linear(450, 200),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Linear(200, n_classes),
            nn.Softmax(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:

        x = self.emb_layer(x)

        # the embedding layer returns the values in format:
        #
        #    (batch_size, sequence_length, embedding_dimension)
        #
        # but the convolution layers expect
        #
        #    (batch_size, num_filters, sequence_length)
        #
        # in order to convolute correctly on the last dimension.
        # this operation swaps the last two elements:
        x = x.permute(0, 2, 1)

        for conv in self.conv_layers:
            x = conv(x)
        # flatten all values
        x = x.view(x.size(0), -1)
        for fc in self.fc_layers:
            x = fc(x)
        return x



from __future__ import annotations

from typing import Any, Dict, Iterable, Optional, Type

import torch
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.preprocessing import LabelEncoder
from skorch import NeuralNetClassifier
from torch import nn


class TorchClassifier(ClassifierMixin, BaseEstimator):
    """
    Simple wrapper for Torch modules enabling string labels and dynamic classes.

    Using this wrapper, the model does not need to know the number of classes that the
    problem has. Instead, the number of classes is learned form the fit data.

    Additionally, this wrapper is able to use string labels by using a LabelEncoder.
    """

    def __init__(
        self,
        module: Type[nn.Module],
        max_seq_len: int = None,
        batch_size: int = 64,
        max_epochs: int = 5,
        learn_rate: float = 1e-3,
        device: str = None,
        model_kwargs: Dict[str, Any] = None,
        optimizer: Type[torch.optim.Optimizer] = torch.optim.Adam,
    ):
        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"

        self.module = module
        self.device = device
        self.batch_size = batch_size
        self.max_epochs = max_epochs
        self.learn_rate = learn_rate
        self.model_kwargs = model_kwargs or {}
        self.wrapped_model: Optional[NeuralNetClassifier] = None
        self.optimizer = optimizer
        self.label_encoder: LabelEncoder = LabelEncoder()
        self.max_seq_len = max_seq_len

    def fit(self, x: Any, y: Iterable[Any], **fit_kwargs: Any) -> TorchClassifier:
        if self.wrapped_model is None:
            classes = set(y)
            n_classes = len(classes)
            self.model_kwargs["module__n_classes"] = n_classes
            self.wrapped_model = NeuralNetClassifier(
                module=self.module,
                device=self.device,
                batch_size=self.batch_size,
                max_epochs=self.max_epochs,
                lr=self.learn_rate,
                optimizer=self.optimizer,
                classes=classes,
                **self.model_kwargs
            )
            y = self.label_encoder.fit_transform(y)
        self.wrapped_model.fit(x, y, **fit_kwargs)
        return self

    def predict(self, x: Any) -> Any:
        if self.wrapped_model is None:
            raise ValueError("model was not fitted")
        predicted_classes = self.wrapped_model.predict(x)
        return self.label_encoder.inverse_transform(predicted_classes)



"""Implementation of Llorens 2016."""
from __future__ import annotations

import random
from collections import defaultdict
from typing import List, Union, Dict

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


def get_features_for_sample(sample: List[str]) -> List[int]:
    """
    Calculate statistical features from samples.

    First, the occurrence of all words is counted.
    Then, this method returns the number of words that are between pre-set
    frequency thresholds.
    Args:
        sample: a list of words to calculate metrics from.

    Returns: a list of four threshold counts: [
            1. total number of words,
            2. number of words that occur once,
            3. number of words that occur 2-4 times,
            4. number of words that occur 5-10 times
        ].

    """
    counts: Dict[str, int] = defaultdict(int)
    for word in sample:
        counts[word] += 1
    v0 = len(counts.keys())
    v1, v2, v3 = 0, 0, 0
    for occurrences in counts.values():
        if occurrences <= 1:
            v1 += 1
        elif occurrences <= 4:
            v2 += 1
        elif occurrences <= 10:
            v3 += 1

    return [v0, v1, v2, v3]


class LifeVectorizer(BaseEstimator, TransformerMixin):
    """Implementation of Llorens 2016."""

    def __init__(
        self,
        fragment_sizes: List[int] = None,
        samples: int = 200,
        sample_type: str = "bow",
        force: bool = True,
    ):
        """
        Initialize the transformer.

        Args:
            fragment_sizes: which sizes to use as window sizes
            samples: how many samples per window size are captured
            sample_type: how the samples are created
            force: if true, calculates samples of too-short texts.
        """
        if fragment_sizes is None:
            fragment_sizes = [200, 500, 800, 1000, 1500, 2000, 3000, 4000]
        valid_sample_types = ["bow", "fragment", "both"]
        if sample_type not in valid_sample_types:
            raise ValueError(
                f"unknown sample type: {sample_type}. valid values: "
                f"{valid_sample_types}"
            )
        self.fragment_sizes = fragment_sizes
        self.samples = samples
        self.sample_type = sample_type
        self.force = force

    def fit(self, _x: List[str], _y: Union[List, np.array] = None) -> LifeVectorizer:
        """Fit the model."""
        return self

    def sample(
        self,
        words: Union[List[str], np.array],
        fragment_size: int,
        method: str,
    ) -> List[List[str]]:
        """
        Take <self.samples> samples of <fragment_size> elements from <words>.

        Args:
            words: a sequence of words to sample from
            fragment_size: size of the window
            method: either "fragment" or "bow" (see paper)

        Returns:
            a list of <self.samples> samples.
        """
        ret = []
        wordcount = len(words)
        if wordcount < fragment_size:
            if self.force:
                fragment_size = wordcount
            else:
                raise ValueError(
                    f"fragment size ({fragment_size}) is larger than document "
                    f"size ({wordcount}) for document starting with: \n\n"
                    f'{" ".join(words[:50])}\n\n'
                )
        for _ in range(self.samples):
            if method == "fragment":
                left = random.randint(0, wordcount - fragment_size)
                right = left + fragment_size
                ret.append(words[left:right])
            if method == "bow":
                ret.append(random.sample(words, fragment_size))
        return ret

    def get_features(
        self,
        document: List[str],
        sample_size: int,
    ) -> np.array:
        """Extract features from a document given a sample size."""
        if self.sample_type == "both":
            return np.concatenate(
                [
                    self._get_features(document, sample_size, "bow"),
                    self._get_features(document, sample_size, "fragment"),
                ]
            )
        else:
            return self._get_features(document, sample_size, self.sample_type)

    def _get_features(
        self,
        document: List[str],
        fragment_size: int,
        method: str,
    ) -> np.array:
        samples = self.sample(document, fragment_size, method)
        features = []
        for sample in samples:
            features.append(get_features_for_sample(sample))
        features = np.array(features)
        means = np.mean(features, axis=0)
        stds = np.std(features, axis=0)
        return np.concatenate(
            [means, np.divide(means, stds, out=np.zeros_like(means), where=stds != 0)]
        )

    def transform(
        self, x: List[List[str]], _y: Union[List, np.array] = None
    ) -> np.array:
        """Calculate samples and extracts features from documents."""
        ret = []
        for document in x:
            doc = [self.get_features(document, size) for size in self.fragment_sizes]
            ret.append(np.concatenate(doc))

        # some classifiers like XGBoost require a numpy array if nested
        return np.array(ret)



"""Transformer computing second-order attributes."""
from __future__ import annotations

import logging
import math
from collections import defaultdict
from typing import Any, Dict, Iterable, List

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

logger = logging.getLogger(__name__)


class SubFrequencyVectorizer(BaseEstimator, TransformerMixin):
    """Transformer computing second-order attributes."""

    def __init__(self) -> None:
        """Initialize the model."""
        self.t: Dict[str, np.array] = defaultdict(np.array)

    def fit(
        self, X: List[str], y: Iterable[str], **_fit_params: Any
    ) -> SubFrequencyVectorizer:
        """Fit data to model."""
        words = set()

        logger.info("starting phase 0")
        documents = defaultdict(list)
        for document, target in zip(X, y):
            documents[target].append(document)
            for word in document.split():
                words.add(word)

        targets = set(y)
        tp: Dict[str, Dict[str, float]] = defaultdict(lambda: defaultdict(float))
        weights: Dict[str, Dict[str, float]] = defaultdict(lambda: defaultdict(float))

        # equation 1: weight calculation
        for target, target_docs in documents.items():
            for document in target_docs:
                doc_words = document.split()
                for doc_word in doc_words:
                    tf = doc_words.count(doc_word)
                    length = len(doc_words)
                    weights[target][doc_word] += math.log2(1 + tf / length)

        # equation 2.2: normalization
        for target in targets:
            for word in words:
                norm_term = 0.0
                for k in targets:
                    norm_term += weights[k][word]
                tp[target][word] = weights[target][word] / norm_term

        for word in words:
            self.t[word] = np.array([tp[target][word] for target in targets])

        return self

    def transform(self, X: Iterable[str], _y: Any = None) -> np.array:
        """Transform data due to previously learned frequencies."""
        result = []
        for k in X:
            document_sum = 0
            doc_words = k.split()
            for j in doc_words:
                if j not in self.t:
                    continue
                tf = doc_words.count(j)
                document_sum += self.t[j] * tf / len(doc_words)
            result.append(document_sum)
        return np.array(result)
